'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _baba = require('./parser/baba');

var _baba2 = _interopRequireDefault(_baba);

var _babelCore = require('babel-core');

var babel = _interopRequireWildcard(_babelCore);

var _babylon = require('babylon');

var babylon = _interopRequireWildcard(_babylon);

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const getAst = (grammar, searchPaths = []) => {
	const getIdentifier = it => `baba$${it.join('$').replace(/[^a-z0-9_]/ig, '_')}`;
	const getFunctionIdentifier = it => `baba$${it.join('$').replace(/[^a-z0-9_]/ig, '_')}$$fn`;
	const arrowWrap = (identifier, arg) => t.callExpression(t.identifier(identifier), [t.arrowFunctionExpression([], arg)]);

	const template = _fs2.default.readFileSync(require.resolve('./templates/default'), 'utf-8');
	const templateAst = babylon.parse(template, { sourceType: 'module' });

	const templateRefs = {};
	(function extractTemplateData(node) {
		if (Array.isArray(node)) {
			node.forEach(sub => extractTemplateData(sub));
			return;
		}

		try {
			const [, m] = node.leadingComments[0].value.trim().match(/@([a-z]+)/);
			if (m) {
				templateRefs[m] = node.declarations[0].id.name;
			}
		} catch (e) {
			// Ignore
		}
	})(templateAst.program.body);

	class DeclarationBlock {
		constructor() {
			this.exports = [];
			this.definitions = [];
			this.vars = [];
			this.mappings = {};
		}

		_getFunctions(node, path = []) {
			let ret = [];
			if (Array.isArray(node)) {
				node.forEach(subNode => {
					ret = ret.concat(this._getFunctions(subNode, path.concat(subNode.key.name || subNode.key.value)));
				});
				return ret;
			}
			if (t.isObjectProperty(node)) {
				if (t.isObjectExpression(node.value)) {
					return ret.concat(this._getFunctions(node.value.properties, path));
				}
				return { path, ast: node.value };
			} else if (t.isObjectMethod(node)) {
				return { path, ast: t.functionExpression(null, node.params, node.body) };
			}
			return ret;
		}

		addImport(file, alias) {
			let contents = null;
			searchPaths.forEach(searchPath => {
				try {
					const filePath = _path2.default.join(searchPath, file + '.baba');
					contents = _fs2.default.readFileSync(filePath, 'utf-8');
					return;
				} catch (e) {
					// ignore
				}
			});
			if (contents === null) {
				throw new Error(`Could not find grammar file "${file}" in any search path:\n\n\t${searchPaths.join(',\n\t')}`);
			}
			reduceTree(_baba2.default.parse(contents), [alias]); // Parse and insert the imported grammar
		}

		addFunction(identifier, args, body) {
			this.definitions.push(t.variableDeclaration('const', [t.variableDeclarator(t.identifier(identifier), t.callExpression(t.identifier(templateRefs.function), [t.arrowFunctionExpression(args.map(arg => t.identifier(arg)),
			// The function must be wrapped in a function body to 
			// be able to parsed, otherwise we'll likely receive 
			// a "return outside function" exception
			babylon.parse(`() => { ${body} }`).program.body[0].expression.body)]))]));
		}

		addMapping(identifier, from, to, type) {
			if (!this.mappings[identifier]) {
				this.mappings[identifier] = [];
			}

			let ret = [from.type === 'regexp' ? babylon.parse(from.value).program.body[0].expression : t.stringLiteral(from.value), t.stringLiteral(to.value)];

			this.mappings[identifier].push(t.arrayExpression(ret));

			if (type === '<->') {
				// Add reverse mapping (bidirectional)
				ret.reverse();
				this.mappings[identifier].push(t.arrayExpression(ret));
			}
		}

		addExport(key, value) {
			this.exports.push(t.objectProperty(t.stringLiteral(key), value));
		}

		addDefinition(identifier, value) {
			this.definitions.push(t.variableDeclaration('const', [t.variableDeclarator(t.identifier(identifier), value)]));
		}

		addVar(identifier, alias, fallback) {
			this.vars.push(t.variableDeclaration('const', [t.variableDeclarator(t.identifier(identifier), t.callExpression(t.identifier(templateRefs.variable), [t.stringLiteral(alias), t.arrowFunctionExpression([], fallback)]))]));
		}

		getAst() {
			let mappings = Object.entries(this.mappings).reduce((acc, [identifier, arr]) => acc.concat(t.variableDeclaration('const', [t.variableDeclarator(t.identifier(identifier), t.callExpression(t.identifier(templateRefs.mapping), arr))])), []);

			return mappings.concat(this.definitions).concat(this.vars).concat(t.exportDefaultDeclaration(t.objectExpression(this.exports)));
		}
	}

	const declarations = new DeclarationBlock();
	const parsed = _baba2.default.parse(grammar);

	let handlers;

	const reduceTree = (nodes, path = []) => {
		// Note: A handler must return an AST node with children, or undefined
		return nodes.map(node => {
			try {
				return handlers[node.type]({ node, path });
			} catch (e) {
				console.error('Error in "%s" at "%s" handler: %s', node.type, path.join('.'), e);
				throw e;
			}
		}).filter(item => !!item); // Remove empty items
	};

	handlers = {
		meta_import({ node }) {
			const [file, alias] = node.arguments;
			declarations.addImport(file.value, alias.value);
		},
		meta_export({ node, path }) {
			const [key, value] = node.arguments;

			declarations.addExport(key.value, t.callExpression(t.identifier(templateRefs.export), [t.arrowFunctionExpression([], reduceTree([value], path)[0])]));
		},
		list_block({ node, path, addDeclaration = true }) {
			const nodePath = node.identifier ? path.concat(node.identifier.value) : [];
			const subTree = reduceTree(node.children, nodePath);

			if (!subTree.length) {
				// Ignore empty scopes/lists
				return null;
			}

			let ret = arrowWrap(templateRefs.choice, t.arrayExpression(subTree));

			if (!nodePath.length) {
				// Anonymous list
				return ret;
			}

			const identifier = getIdentifier(nodePath);

			if (addDeclaration) {
				if (node.identifier.type === 'var_identifier') {
					// Variable declaration and fallback definition
					declarations.addVar(identifier, node.identifier.value, ret);
				} else {
					declarations.addDefinition(identifier, ret);
				}
			}

			return t.identifier(identifier);
		},
		scope_block({ node, path }) {
			// TODO Scopes may be exported if declarations are added, make this 
			// configurable. It's off by default to save space in the generated 
			// JS file.
			return this.list_block({ node, path, addDeclaration: false });
		},
		tag_block({ node, path }) {
			const subTree = reduceTree([node.tag], path);
			const identifier = getIdentifier(path.concat(node.identifier.value));

			let ret = arrowWrap(templateRefs.choice, t.arrayExpression(subTree));

			if (node.identifier.type === 'var_identifier') {
				// Variable declaration and fallback definition
				declarations.addVar(identifier, node.identifier.value, ret);
			} else {
				declarations.addDefinition(identifier, ret);
			}

			return t.identifier(identifier);
		},
		quoted_string_block({ node, path }) {
			const subTree = reduceTree([node.string], path);
			const identifier = getIdentifier(path.concat(node.identifier.value));

			let ret = subTree[0];

			if (node.identifier.type === 'var_identifier') {
				// Variable declaration and fallback definition
				declarations.addVar(identifier, node.identifier.value, ret);
			} else {
				declarations.addDefinition(identifier, ret);
			}

			return t.identifier(identifier);
		},
		literal({ node }) {
			return t.stringLiteral(node.value);
		},
		function({ node, path }) {
			declarations.addFunction(getFunctionIdentifier(path.concat(node.identifier.value)), node.arguments.map(arg => arg.value), node.body);
		},
		mapping({ node, path }) {
			declarations.addMapping(getFunctionIdentifier(path), node.from, node.to, node.dir);
		},
		identifier({ node, path }) {
			return t.identifier(getIdentifier(path.concat(node.value.split('.'))));
		},
		var_identifier({ node, path }) {
			return this.identifier({ node, path });
		},
		var_assign({ node }) {
			// Returns `VARIABLE.a(() => VALUE`
			return t.callExpression(t.memberExpression(t.identifier(getIdentifier(node.name.value.split('.'))), t.identifier('a')), [t.arrowFunctionExpression([], reduceTree(node.value)[0])]);
		},
		function_identifier({ node, path }) {
			return t.identifier(getFunctionIdentifier(path.concat(node.value.split('.'))));
		},
		interpolated_string({ node }) {
			let subTree = reduceTree(node.children);
			let ret = arrowWrap(templateRefs.concat, t.arrayExpression(subTree));
			if (node.weight > 1) {
				// Returns `new Array(weight).fill(CONCAT(() => NODES))`
				return t.callExpression(t.memberExpression(t.newExpression(t.identifier('Array'), [t.numericLiteral(node.weight)]), t.identifier('fill')), [ret]);
			}
			// Returns `CONCAT(() => NODES)`
			return ret;
		},
		tag({ node, path }) {
			if (node.quantifier === '?') {
				// "Optional" quantifier
				node.children.push({ type: 'literal', value: '' });
				return this.tag_choice({ node, path });
			}
			return reduceTree(node.children, path)[0];
		},
		tag_choice({ node, path }) {
			return this.list_block({ node, path });
		},
		tag_concat({ node, path }) {
			return arrowWrap(templateRefs.concat, t.arrayExpression(reduceTree(node.children, path)));
		},
		transform({ node, path }) {
			const argTree = reduceTree(node.args, path);
			const fnTree = reduceTree(node.fn, path);
			return t.callExpression(fnTree[0], [argTree[0]]);
		},
		function_call({ node, path }) {
			// Note: Function calls must return a callable
			const argTree = reduceTree(node.args, path);
			const fnIdentifier = t.identifier(getFunctionIdentifier(node.fn.value.split('.')));
			return t.callExpression(fnIdentifier, argTree);
		}
	};

	reduceTree(parsed);

	templateAst.program.body = templateAst.program.body.concat(declarations.getAst());

	return templateAst;
};

exports.default = (file, targets, minify = false) => {
	const presets = [[require('babel-preset-env'), { targets }]];

	if (minify) {
		presets.push([require('babel-preset-minify'), {
			mangle: {
				topLevel: true
			}
		}]);
	}

	const filePath = _path2.default.resolve(_path2.default.dirname(file));
	const searchPaths = [filePath, _path2.default.join(filePath, 'node_modules')];

	return babel.transformFromAst(getAst(_fs2.default.readFileSync(file, 'utf-8'), searchPaths), null, {
		presets,
		babelrc: false,
		generatorOpts: {
			comments: false
		}
	}).code;
};